import napari
import numpy as np
from qtpy.QtWidgets import QWidget, QVBoxLayout
from qtpy.QtCore import Qt # type: ignore
from qtpy import QtGui
import fish_feats.MainImage as mi
import fish_feats.Configuration as cf
import fish_feats.Utils as ut
from fish_feats.NapaRNA import NapaRNA, OverlapRNA
from fish_feats.NapaCells import MainCells, Position3D, EndCells 
from fish_feats.NapaNuclei import MeasureNuclei, NucleiWidget, PreprocessNuclei 
from fish_feats.FishGrid import FishGrid
from fish_feats.NapaMix import CheckScale, CropImage, Association, Separation, CytoplasmMeasure
from fish_feats import ClassifyCells as cc
import fish_feats.FishWidgets as fwid
from fish_feats._button_grid import ButtonGrid

"""
    Handle the UI through napari plugin

    Fish&Feats proposes several actions from cell/nuclei segmentation, association, mRNA-Fish segmentation/association and quantitative measurements.

    Available under BSD License
    If you use the code or part of it, please cite associated work.

    Author: Gaëlle Letort, DSCB, Institut Pasteur/CNRS
"""

def show_documentation():
    """ Open the documentation page """
    ut.show_documentation_page("")
    return

## start without viewer for tests
def initZen():
    """ Initialize the plugin with the current viewer """
    ffeats = FishFeats()
    ffeats.init_viewer()

def startZen():
    """ Start the pipeline: open the image, get the scaling infos """
    ffeats = FishFeats()
    ffeats.init_viewer()
    return ffeats.open()

def startFromLayers():
    """ Starts the plugin on already opened image """
    ffeats = FishFeats()
    ffeats.init_viewer()
    if ffeats.viewer is None:
        ut.show_error("No viewer found")
        return
    return ffeats.openFromLayers()

def startMultiscale():
    """ Starts with multiscale view """
    ffeats = FishFeats()
    ffeats.init_viewer()
    return ffeats.startMultiscale()
    
def convert_previous_results():
    """ Convert the previous results to the new format """
    filename = ut.dialog_filename()
    if filename is None:
        print("No file selected")
        return
    ffeats = FishFeats()
    return ffeats.convert_previous_results( filename )
    

def unremove(layer):
    ut.show_info("Removing layer locked, throw an error ")
    #print(str(error))
    return

class FishFeats:
    """ Handle main interface of fishfeats """

    def __init__( self, viewer=None ):
        self.viewer = viewer
        self.cfg = None
        self.mig = None
        self.main_wid = None

        if self.viewer is None:
            self.viewer = napari.current_viewer()
            self.viewer.title = "Fish&Feats"


    #### Start
    def init_viewer( self ):
        """ Launch the plugin, initialize all """
        self.mig = mi.MainImage( talkative=True )
        self.my_cmap = ut.create_labelmap()
        self.persp = 45
        self.viewer.scale_bar.visible = True

        @self.viewer.bind_key('h', overwrite=True)
        def show_help(layer):
            ut.showHideOverlayText(self.viewer)

        @self.viewer.bind_key('F1', overwrite=True)
        def show_layer(viewer):
            self.show_hide( 0 )

        @self.viewer.bind_key('F2', overwrite=True)
        def show_layer(viewer):
            self.show_hide( 1 )

        @self.viewer.bind_key('F3', overwrite=True)
        def show_layer(viewer):
            self.show_hide( 2 )

        @self.viewer.bind_key('F4', overwrite=True)
        def show_layer(viewer):
            self.show_hide( 3 )

        @self.viewer.bind_key('F5', overwrite=True)
        def show_layer(viewer):
            self.show_hide( 4 )

        @self.viewer.bind_key('F6', overwrite=True)
        def show_layer(viewer):
            self.show_hide( 5 )

        @self.viewer.bind_key('F7', overwrite=True)
        def show_layer(viewer):
            self.show_hide( 6 )

        @self.viewer.bind_key('F8', overwrite=True)
        def show_layer(viewer):
            self.show_hide( 7 )

        def show_hide( self, intlayer ):
            """ Show/hide the ith-layer """
            if 0 <= intlayer < len( self.viewer.layers ):
                self.viewer.layers[intlayer].visible = not self.viewer.layers[intlayer].visible

        @self.viewer.bind_key('Ctrl-h', overwrite=True)
        def show_shortcuts(layer):
            ut.main_shortcuts(self.viewer)

        @self.viewer.bind_key('g', overwrite=True)
        def show_grid(layer):
            self.addGrid()

        @self.viewer.bind_key('Ctrl-v', overwrite=True)
        def set_vispymode(viewer):
            pers = self.viewer.camera.perspective
            if pers > 0:
                self.persp = pers
                self.viewer.camera.perspective = 0
            else:
                self.viewer.camera.perspective = self.persp

    def open( self ):
        """ Start the pipeline: open the image, get the scaling infos """

        ## get and open the image
        filename = ut.dialog_filename()
        if filename is None:
            print("No file selected")
            return
        ut.showOverlayText(self.viewer,  "Opening image...")
        self.mig.open_image( filename=filename )
        ut.update_history(self.mig.imagedir)
        self.cfg = cf.Configuration( self.mig.save_filename(), show=False )

        ## display the different channels
        self.display_channels()
        self.endInit()
        return self.main_wid

    def openFromLayers( self ):
        """ Open from already opened layers in napari """
        ut.show_info("Loading all opened layers as channels of one image in FishFeats...")
        self.mig = mi.MainImage(talkative=True)
        if len(self.viewer.layers) == 0:
            ut.show_error("No layer(s) found")
            return None
        scale = self.viewer.layers[0].scale
        imshape = self.viewer.layers[0].data.shape
        self.mig.set_scales(scale[0], scale[1])

        ## single layer with all the channels
        if len(self.viewer.layers[0].data.shape) == 4:
            img = self.viewer.layers[0].data
            img = ut.arrange_dims( img, verbose=True )
            self.mig.set_image( img )
            ut.remove_all_layers( self.viewer )
            self.display_channels()
            return self.getImagePath()
        
        ## Or load all opened layer in the image and rename them in FishFeats style
        img = [] 
        for lay in self.viewer.layers:
            if len(lay.data.shape) == 3:
                if lay.data.shape != imshape:
                    ut.show_error("All layers should have the same shape")
                    return
                img.append( lay.data )
        self.mig.set_image(img)
        ut.remove_all_layers( self.viewer )
        self.display_channels()
        return self.getImagePath()

    def convert_previous_results(self, filename):
        """ Convert the previous results to the new format """
        self.mig = mi.MainImage( talkative=True )
        self.mig.set_imagename( filename )
    
        ## read the config file to extract the scaling and the direction
        self.cfg = cf.Configuration(self.mig.save_filename(), show=False)
        if self.cfg.has_config():
            self.cfg.read_scale(self.mig)

        ## load cell file
        results_filename = self.mig.get_filename( endname = "_results.csv", ifexist=True )
        if results_filename != "":
            self.mig.load_from_results( results_filename )
        else:
            loadfilename = self.mig.junction_filename(dim=2,ifexist=True)
            if loadfilename != "":
                print("Load junctions from file "+loadfilename)
                self.mig.loadCellsFromSegmentation( loadfilename )

            nucleifilename = self.mig.nuclei_filename(ifexist=True)
            if nucleifilename != "":
                print("Load nuclei from file "+str(nucleifilename))
                self.mig.load_segmentation_nuclei(nucleifilename, load_stain=False)
                self.mig.popNucleiFromMask()

        ## Load the RNA files if found some
        for chan in range(30):
            ## image was not read so todn't know how many channels are possible
            ## if find RNA file, load it
            rnafile = self.mig.rna_filename(chan=chan, how=".csv", ifexist=True)
            if rnafile == "":
                rnafile = self.mig.rna_filename(chan=chan, how=".tif", ifexist=True)
            if rnafile != "":
                self.mig.load_rnafile(rnafile, chan, topop=True)

        ## Load cytoplasmic results file if found some
        cytofile = self.mig.get_filename(endname="_cytoplasmic.csv", ifexist=True)
        if cytofile != "":
            self.mig.loadCytoplasmicTable( cytofile )
    
        ## Load features files if found some
        featfile = self.mig.get_filename(endname="_features.csv", ifexist=True)
        if featfile != "":
            self.mig.loadFeatureTable( featfile )

        ## Save the new results file
        self.mig.save_results()
        return QWidget() 

    def display_channels(self):
        """ Display the different channels of the image """
        if self.viewer is None or self.mig is None:
            return
        cmaps = ut.colormaps()
        ncmaps = len(cmaps)
        for channel in range(self.mig.nbchannels):
            cmap = cmaps[(channel%ncmaps)]
            img = self.mig.get_channel(channel)
            self.viewer.add_image( img, name="originalChannel"+str(channel), blending="additive", scale=(self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY), colormap=cmap, contrast_limits=ut.quantiles(img), gamma=0.9 )
        self.viewer.axes.visible = True

    def endInit(self):
        """ Finish, go to metadata step """
        self.viewer.grid.enabled = True
        ut.removeOverlayText(self.viewer)
        self.checkScale()
        return self.main_wid

    def shortcuts_window(self):
        """ Open a separate text window with the main steps and shortcuts """
        vie = napari.current_viewer()
        ut.main_shortcuts(vie)


    def getImagePath(self):
        """ Get the image path when it was open from layers """

        image_path = fwid.file_dialog( "Select image path", "All files (*)", directory=self.mig.get_image_path() )

        self.mig.set_image_path(image_path)
        ut.update_history(self.mig.imagedir)
        if self.cfg is None:
            self.cfg = cf.Configuration(self.mig.save_filename(), show=False)
        self.endInit()
        return self.main_wid


    ### Grid tools: regular grid for spatial repere
    def addGrid(self):
        """ Interface to create/load a grid for repere """
        if "FishGrid" not in self.viewer.layers:
            grid = FishGrid(self.viewer, self.mig)
            self.viewer.window.add_dock_widget(grid, name="FishGrid")
        else:
            gridlay = self.viewer.layers["FishGrid"]
            gridlay.visible = not gridlay.visible


    def startMultiscale(self):
        """ Open the main image as multiscale for performance """
        filename = ut.dialog_filename()
        if filename is None:
            ut.show_error("No file selected, try again")
            return
        ut.showOverlayText(self.viewer,  "Opening image...")
        self.mig.open_image( filename=filename )
        ut.update_history(self.mig.imagedir)
        self.cfg = cf.Configuration(self.mig.save_filename(), show=False)
    
        for channel in range(self.mig.nbchannels):
            cmap = ut.colormapname(channel)
            img = self.mig.get_channel(channel)
            cview = self.viewer.add_image( [img, img[:,::2,::2], img[:,::4,::4] ], name="originalChannel"+str(channel), blending="additive", scale=(self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY), colormap=cmap )
            cview.contrast_limits=ut.quantiles(img)
            cview.gamma=0.95
        self.viewer.axes.visible = True
        self.endInit()
        return self.main_wid

    def byebye( self ):
        """ Quit the pipeline """
        ut.remove_all_layers( self.viewer )
        self.viewer.title = "napari"
        ut.removeOverlayText( self.viewer )
        if self.cfg.blabla.shown():
            self.cfg.blabla.close()
        self.cfg.write_parameterfile()
        #ut.remove_widget( viewer, "Main" )
        print("Bye bye")
        #del mig
        try:
            self.main_wid.close()
            ut.remove_all_widget( self.viewer ) 
        except:
            print("Partial closing")
        #ut.remove_all_widget( self.viewer ) 
        del self.mig
        del self.cfg

    def checkScale(self):
        """ Interface to choose the image scales and channels """
        cs = CheckScale(self)
        wid = self.viewer.window.add_dock_widget(cs, name="Scale")
        self.cfg = cs.cfg
        self.main_wid = wid

    #### Action choice
    def getChoices( self, default_action='Cells:Segment'):
        """ Launch the interface of Main step """
        choices = {}

        choices["Init:Separate stainings"] = self.divorceJunctionsNuclei
        # actions related to cell contour
        choices['Cells:Segment'] = self.goJunctions
        choices['Cells:3D position'] = self.show3DCells
        # actions related to nuclei
        choices['Nuclei:Segment'] = self.getNuclei
        choices["Nuclei:Associate to cells"] = self.doCellAssociation

        choices['Nuclei:Preprocess'] = self.preprocNuclei
        # related to RNA
        choices['RNA:Segment&assign'] = self.getRNA
        choices['RNA:Get overlaps'] = self.getOverlapRNA
        # Analysis options
        choices['Measure:Cytoplasmic intensity'] = self.cytoplasmicStaining
        choices['Measure:Nuclear intensity'] = self.measureNuclearIntensity
        choices['Measure:Classify cells'] = self.launch_classify
        # Utilities options
        choices['Misc:Quit plugin'] = self.byebye
        choices['Misc:Image scalings'  ] = self.checkScale
        choices['Misc:Touching labels'] = self.touching_labels
        choices['Misc:Add grid'] = self.addGrid 
        choices['Misc:Crop image'] = self.crop_image 

        choice_wid = GetChoices( default_action, choices, self.cfg )
        ut.remove_widget( self.viewer, "Main" )
        wid = self.viewer.window.add_dock_widget(choice_wid, name="Main")
        self.main_wid = wid

    def launch_classify( self ):
        """ Launch the cell classification """
        if not self.mig.hasCells():
            ut.show_info("No cells - segment/load it before")
        else:
            cc.classify_cells(self.mig, self.viewer)

    def crop_image(self):
        """ Interface to crop the image and associated segmentations/results """
        crop = CropImage(self)
        self.viewer.window.add_dock_widget( crop, name="CropImage" )

    def load_all_previous_files(self):
        """ Load all the previous files with default name that it can find and init the objects accordingly """
        ## try to load separated staining
        if self.mig.should_separate():
            separated_junctionsfile = self.mig.separated_junctions_filename(ifexist=True)
            separated_nucleifile = self.mig.separated_nuclei_filename(ifexist=True)
            if (separated_junctionsfile != "") and (separated_nucleifile != ""):
                self.load_separated( separated_junctionsfile, separated_nucleifile )
    
        loadfilename = self.mig.junction_filename(dim=2,ifexist=True)
        if loadfilename != "":
            self.cfg.addText("Load junctions from file "+loadfilename)
            self.mig.load_segmentation( loadfilename )
            self.mig.popFromJunctions()

        nucleifilename = self.mig.nuclei_filename(ifexist=True)
        if nucleifilename != "":
            self.cfg.addText("Load nuclei from file "+str(nucleifilename))
            self.mig.load_segmentation_nuclei(nucleifilename)
            self.mig.popNucleiFromMask()

        for chan in self.mig.potential_rnas():
            ## if find RNA file, load it
            rnafile = self.mig.rna_filename(chan=chan, how=".csv", ifexist=True)
            if rnafile == "":
                rnafile = self.mig.rna_filename(chan=chan, how=".tif", ifexist=True)
            if rnafile != "":
                self.mig.load_rnafile(rnafile, chan, topop=True)

        ut.remove_all_widget( self.viewer )
        self.getChoices(default_action="Classify cells")

    def load_separated( self, juncfile, nucfile ):
        """ load the separated files """
        self.mig.load_separated_staining( juncfile, nucfile )
        ut.show_info("Separated stainings loaded")

    def loadJunctionsFile(self):
        """ Load the segmentation from given file and directly init the cells """
        loadfilename = self.mig.junction_filename(dim=2,ifexist=True)
        self.cfg.addText("Load junctions from file "+loadfilename)
        self.mig.load_segmentation( loadfilename )
        self.mig.popFromJunctions()
        ut.remove_all_widget( self.viewer )
        self.getChoices(default_action="Get nuclei")

    def goJunctions(self):
        """ Choose between loading projection and cells files or recalculating """
        if self.mig.junchan is None:
            ut.show_warning( "No junction channel selected in the configuration. Go back to Image Scalings to select one." )
            return

        main_cells = MainCells( self ) 
        if main_cells.proj is not None:
            self.viewer.window.add_dock_widget( main_cells.proj, name="JunctionProjection2D" )
        else:
            self.viewer.window.add_dock_widget( main_cells, name="Get cells" )


    ################################
    ###### Show cell in 3D and possibility to edit the Z position of cells
    def show3DCells(self):
        """ Cells in 3D and update Z position of cells """
        cells_3D = Position3D( self )
        self.viewer.window.add_dock_widget(cells_3D, name="Cells in 3D")

    ##### preprocessing functions
    def preprocNuclei(self):
        """ Preprocess the nuclei before segmentation """
        preproc_nuclei = PreprocessNuclei( self.viewer, self.mig, self.cfg )
        self.viewer.window.add_dock_widget(preproc_nuclei, name="Preprocess Nuclei")

    ################### Junction and nuclei separation functions

    def divorceJunctionsNuclei(self):
        """ Separate the junctions and nuclei staining if they are in the same channel """
        separation = Separation( self ) 
        self.viewer.window.add_dock_widget(separation, name="Separate")

    ################ Nuclei segmentation
    def getNuclei(self):
        """ 3D segmentation and correction of nuclei """
        print("******* 3D segmentation of nuclei ******")
        text = "Choose method and parameters to segment nuclei in 3D \n"
        text += "The nuclei are segmented from the original nuclei channel if the stainings are separate \n"
        text += "Or from the nucleiStaining image if the staining were originally mixed \n"
        ut.showOverlayText(self.viewer, text)

        ut.hide_color_layers(self.viewer, self.mig)
        ut.show_layer(self.viewer, self.mig.nucchan)
        
        nuclei_widget = NucleiWidget( self )
        self.viewer.window.add_dock_widget( nuclei_widget, name="Get nuclei" )

    ######################### Association of 2D cells with nuclei
    def doCellAssociation(self):
        """ Association of nuclei with corresponding apical junction cells """
        do_association = Association( self.viewer, self.mig, self.cfg )
        self.viewer.window.add_dock_widget(do_association, name="Associating")

    ######################### RNA
    def getOverlapRNA(self):
        """ Find RNAs overlapping in several channels (non specific signal) """
        over = OverlapRNA(self.viewer, self.mig, self.cfg)
        self.viewer.window.add_dock_widget( over, name="Overlapping RNAs" )

    def getRNA(self):
        """ Segment the RNA dots in selected channels """
        if not ut.has_widget( self.viewer, "RNAs"):
            rnaGUI = NapaRNA(self.viewer, self.mig, self.cfg)


    ######## Labels edition
    def showCellsWidget(self, layerName, shapeName='CellNames', dim=3):
        """ option to show and edit cell segmentation """
        layer = self.viewer.layers[layerName]
        
        @layer.bind_key('Control-c', overwrite=True)
        def contour_increase(layer):
            if layer is not None:
                layer.contour = layer.contour + 1
        
        @layer.bind_key('Control-d', overwrite=True)
        def contour_decrease(layer):
            if layer is not None:
                if layer.contour > 0:
                    layer.contour = layer.contour - 1
        
    
        @layer.mouse_drag_callbacks.append
        def clicks_label(layer, event):
            if event.type == "mouse_press":
                if len(event.modifiers) == 0:
                    if event.button == 2:
                        label = layer.get_value(position=event.position, view_direction = event.view_direction, dims_displayed=event.dims_displayed, world=True)
                        layer.selected_label = label
                        layer.refresh()
                    return
        
                if 'Control' in event.modifiers:
                    if event.button == 2:
                        ### Erase a label
                        label = layer.get_value(position=event.position, view_direction = event.view_direction, dims_displayed=event.dims_displayed, world=True)
                        layer.data[layer.data==label] = 0
                        layer.refresh()
                        return

                    if event.button == 1:
                        ## Merge two labels
                        start_label = layer.get_value(position=event.position, view_direction = event.view_direction, dims_displayed=event.dims_displayed, world=True)
                        yield
                        while event.type == 'mouse_move':
                            yield
                        end_label = layer.get_value(position=event.position, view_direction = event.view_direction, dims_displayed=event.dims_displayed, world=True)
                        # Control left-click: merge labels at each end of the click
                        print("Merge label "+str(start_label)+" with "+str(end_label))
                        frame = None
                        if layer.ndim == 3:
                            frame = int(self.viewer.dims.current_step[0])
                        ut.merge_labels( layer, frame, start_label, end_label )
    

        @layer.bind_key('m', overwrite=True)
        def set_maxlabel(layer):
            mess = "Max label on image "+shapeName+": "+str(np.max(layer.data)) 
            mess += "\n "
            mess += "Number of labels used: "+str(len(np.unique(layer.data)))
            ut.show_info( mess )
            layer.mode = "PAINT"
            layer.selected_label = np.max(layer.data)+1
            if layer.selected_label == 1:
                layer.selected_label = 2
            layer.refresh()
            return

        @layer.bind_key('l', overwrite=True)
        def switch_show_lab(layer):
            if shapeName in self.viewer.layers:
                self.viewer.layers.remove(shapeName)
            else:
                ut.get_bblayer(layer, shapeName, dim, self.viewer, self.mig)
            return
    
        def relabel_layer():
            maxlab = np.max(layer.data)
            used = np.unique(layer.data)
            nlabs = len(used)
            if nlabs == maxlab:
                print("already relabelled")
                return
            for j in range(2, nlabs+1):
                if j not in used:
                    layer.data[layer.data==maxlab] = j
                    maxlab = np.max(layer.data)
            layer.refresh()

        def show_names():
            if shapeName in self.viewer.layers:
                self.viewer.layers.remove(shapeName)

        help_text = ut.labels_shortcuts( level = 0 )
        header = ut.helpHeader(self.viewer, layerName)
        ut.showOverlayText(self.viewer, header+help_text)
    
        print( "\n #########################################\n Labels correction options:\n " + help_text + self.textCellsWidget() )

        if "Junctions" in self.viewer.layers:
            self.viewer.layers["Junctions"].preserve_labels = True

    def textCellsWidget(self):
        text = "  <Control+left click> from one label to another to merge them (the label kept will be the last one) \n"
        text += "'show_cellnames' (<l>) add a new layer showing the label (number) around each object position. \n"
        #text += "'relabel update' the cell labels to have consecutives numbers from 2 to number_of_cells.\n"
        text += "\n For 3D: \n"
        text += "In 3D, most label actions wont work if Vispy perspective is ON. Switch it off with 'Ctrl-v' before.\n"
        text += "If n_edit_dim is set on 3 (top left panel), edition will affect all or several z (slices) \n"
        text += "If n_edit_dim is set on 2, edition will only affect the active slice \n"
        return text
    
    def correction_junctions( self ):
        """ Manual correction of segmentation step """
        ut.removeOverlayText( self.viewer )
        ut.remove_widget( self.viewer, "Get cells" )
        maskview = self.viewer.add_labels( self.mig.junmask, blending='additive', scale=(self.mig.scaleXY, self.mig.scaleXY), name="Junctions" )
        maskview.contour = 3
        maskview.selected_label = 2
        self.showCellsWidget( "Junctions", shapeName="JunctionsName", dim=2 )
        #ut.remove_widget( self.viewer, "Get cells" )
        # saving and finishing
        endcells = EndCells( self )
        self.viewer.window.add_dock_widget( endcells, name="EndCells" )


    def measureNuclearIntensity(self):
        """ Measure intensity inside segmented nuclei """
        if not self.mig.hasNuclei():
            ut.show_warning( "Segment/Load nuclei before" )
            return
        meas_nuc = MeasureNuclei( self.viewer, self.mig, self.cfg )
        self.viewer.window.add_dock_widget(meas_nuc, name="Measure nuclei")

    ############ measure cyto
    def cytoplasmicStaining(self):
        """ Measure the cytoplasmic signal close to the apical surface """
        cytoMeas = CytoplasmMeasure( self.viewer, self.mig, self.cfg )
        self.viewer.window.add_dock_widget(cytoMeas, name="Measure cytos")

    def helpMessageEditContours(self, dim=2):
        text = '- To see the cells as filled area, put 0 in the *contour* field. Else to see the contours, in *contour* field, put 1 or more (will be thicker if >1) \n'
        text = text + '- To erase one label entirely, put *0* in the *label* field, select the *fill* tool and click on the label \n'
        text = text + '- To add one label, choose a label value higher than all the ones in the image,'
        text = text + ' put it in the *label* field, select the *drawing* tool and draw it.'
        text = text + ' Fill the contour you have drawn to finish the new cell \n'
        text = text + '- To draw, choose the label value in the *label* field, and click and drag to paint.'
        text += ' If *preserve labels* is selected, drawing above another label doesn t affect it \n'
        text += '- Holding space to zoom/unzoom \n'
        #text += 'You can set the selected label to be one larger than the current largest label by pressing M.\n'
        if dim == 3:
            text += '- Check n_edit_dimensions box to modify all 3D nuclei at once (else work only on the current slice) \n'
        return text

############################## Extra-tools

    ### Touching labels for Griottes
    def touching_labels(self):
        """ Dilate labels so that they all touch """
        ## perform the label expansion
        from skimage.morphology import binary_opening
        from skimage.segmentation import expand_labels
        print("********** Generate touching labels image ***********")
    
        ## get junctions img
        if "Cells" not in self.viewer.layers:
            if self.mig.pop is None or self.mig.pop.imgcell is None:
                ut.show_info("Load segmentation before!")
                return
            labimg = self.viewer.add_labels(self.mig.pop.imgcell, name="Cells", scale=(self.mig.scaleXY, self.mig.scaleXY), opacity=1, blending="additive")
            labimg.contour = 0

        ## skeletonize it
        img = self.viewer.layers["Cells"].data
        ext = np.zeros(img.shape, dtype="uint8")
        ext[img==0] = 1
        ext = binary_opening(ext, footprint=np.ones((2,2)))
        newimg = expand_labels(img, distance=4)
        newimg[ext>0] = 0
        newlay = self.viewer.add_labels(newimg, name="TouchingCells", scale=(self.mig.scaleXY, self.mig.scaleXY), opacity=1, blending="additive")
        newlay.contour = 0

        ## open the widget for options
        tlabel_wid = TouchingLabels( self.viewer, self.mig, self.cfg )
        self.viewer.window.add_dock_widget(tlabel_wid, name="Touching labels")

class TouchingLabels( QWidget):
    """ Generate an image with touching labels from the junctions image, handle compability with Griottes """
    def __init__(self, viewer, mig, cfg):
        super().__init__()
        self.viewer = viewer
        self.mig = mig
        self.cfg = cfg

        layout = QVBoxLayout()
        ## save the image
        save_btn = fwid.add_button( "Save touching labels image", self.save_touching_labels_image, descr="Save the resulting image of expanded cell labels", color=ut.get_color("save") )  
        layout.addWidget( save_btn )
        
        ## if Griottes has run, adapt it to the scale
        scale_btn = fwid.add_button( "Scale Griottes image", self.scale_griottes, descr="Scale the images resulting from Griottes computing to the main image scale" )  
        layout.addWidget( scale_btn )
        self.setLayout(layout)
    
    def save_touching_labels_image( self ):
        """ Save the touching labels image """
        if "TouchingCells" not in self.viewer.layers:
            ut.show_error("No touching labels image to save")
            return
        outname = self.mig.build_filename( "_touching_labels.tif")
        self.mig.save_image(self.viewer.layers["TouchingCells"].data, imagename=outname)
        ut.show_info("Saved touching labels image as "+outname)

    def scale_griottes( self ):
        """ Scale the Griottes images to the main image scale """
        ut.scale_layer( self.viewer, "Centers", (self.mig.scaleXY, self.mig.scaleXY) )
        ut.scale_layer( self.viewer, "Contact graph", (self.mig.scaleXY, self.mig.scaleXY) )
        ut.scale_layer( self.viewer, "Graph", (self.mig.scaleXY, self.mig.scaleXY) )


def test():
    print("for test")


class GetChoices( QWidget ):
    """ Main widget with all the action choices """

    def __init__( self, default_action, actions, cfg ):
        """ Initialiaze the Main interface with the choice of action to perform """
        super().__init__()
        self.default_action = default_action
        self.actions_functions = actions
        self.cfg = cfg

        layout = QVBoxLayout()
        ## Choice list
        action_line, self.action = fwid.list_line( "Action: ", descr="Choose which action (step) to perform now", func=None )
        choices = self.actions_functions.keys()
        for choice in choices:
            self.action.addItem(choice)
        self.action.setCurrentText( self.default_action )

        ## Set the colors of the steps
        font = QtGui.QFont()
        font.setItalic(True)
        for ind in range(self.action.count()):
            step = self.action.itemText( ind )
            #if step.startswith( "Init:" ):
            #    self.action.setItemData( ind, QtGui.QColor("#514B64"), Qt.BackgroundRole )
            if step.startswith( "Cells:" ):
                self.action.setItemData( ind, QtGui.QColor("#AA2626"), Qt.BackgroundRole )
            if step.startswith( "Nuclei:" ):
                self.action.setItemData( ind, QtGui.QColor("#214097"), Qt.BackgroundRole )
            if step.startswith( "RNA:" ):
                self.action.setItemData( ind, QtGui.QColor("#197A30"), Qt.BackgroundRole )
            if step.startswith( "Measure:" ):
                self.action.setItemData( ind, QtGui.QColor("#B3971B"), Qt.BackgroundRole )
            if step.startswith( "Misc:" ):
                self.action.setItemData( ind, QtGui.QColor("#969694"), Qt.BackgroundRole )
            if step.startswith("Nuclei:Preprocess"):
                self.action.setItemData( ind, font, Qt.FontRole )
            if step.startswith("RNA:Get overlaps"):
                self.action.setItemData( ind, font, Qt.FontRole )

        layout.addLayout( action_line )
        ## button go
        go_btn = fwid.add_button( "GO", self.launch_action, descr="Launch the selected action", color=ut.get_color("go") )
        layout.addWidget(go_btn)

        self.setLayout(layout)
        self.action.currentTextChanged.connect( self.launch_action )

    def launch_action(self):
        """ Launch next step with selected action """
        action = self.action.currentText()
        self.cfg.addSectionText(action)
        self.actions_functions[action]()
    




import fish_feats.Utils as ut
import fish_feats.FishWidgets as fwid
import fish_feats.Configuration as cf
import numpy as np
import os
from qtpy.QtWidgets import QVBoxLayout, QWidget 

class CheckScale( QWidget):
    """
        Handle the update of metadata parameters, choice of channel
    """

    def __init__( self, ffeats ):
        """
            Initialize the interface to set metadata
        """
        self.viewer = ffeats.viewer
        self.mig = ffeats.mig
        self.cfg = ffeats.cfg
        self.ffeats = ffeats
        super().__init__()
        layout = QVBoxLayout()

        if self.cfg is None:
            self.cfg = cf.Configuration(self.mig.save_filename(), show=False)

        ## load saved parameters
        if self.cfg.has_config():
            self.cfg.read_scale(self.mig)
        zdir = "top high z"
        if self.mig.zdirection == 1:
            zdir = "top low z"
        
        ## get the scaling in XY
        line_scalexy, self.scaleXY = fwid.value_line( "XY scale (um/pixel):", self.mig.scaleXY, descr="Set the scale in XY (um/pixel)" )
        layout.addLayout( line_scalexy )
        
        ## get the scaling in Z
        line_scalez, self.scaleZ = fwid.value_line( "Z scale (um/pixel):", self.mig.scaleZ, descr="Set the scale in Z (um/pixel)" )
        layout.addLayout( line_scalez )

        ## get the direction of Z
        line_zdir, self.zdirection = fwid.list_line( "Z direction:", descr="Choose the direction of Z" )
        layout.addLayout( line_zdir )
        self.zdirection.addItems( ["top high z", "top low z"] )
        self.zdirection.setCurrentText( zdir )

        ## get the channel number of junction if any
        line_junchan, self.junchan = fwid.list_line( "Junction channel:", descr="Choose the channel number of the junction staining, or None if no junction staining" )
        layout.addLayout( line_junchan )
        self.junchan.addItem( "None" )
        for chan in range(self.mig.nbchannels):
            self.junchan.addItem( str(chan) )
        self.junchan.setCurrentText( str(self.mig.junchan) )

        ## get the channel number of nuclei if any
        line_nucchan, self.nucchan = fwid.list_line( "Nuclei channel:", descr="Choose the channel number of the nuclei staining, or None if no nuclei staining" )
        layout.addLayout( line_nucchan )
        self.nucchan.addItem( "None" )
        for chan in range(self.mig.nbchannels):
            self.nucchan.addItem( str(chan) )
        self.nucchan.setCurrentText( str(self.mig.nucchan) )

        ## load previous results
        self.load_previous = fwid.add_check( "Load previous", True, None, descr="Load previously saved files (in results folder) for the current image" )
        layout.addWidget( self.load_previous )

        ## button to open help
        help_btn = fwid.add_button( "Help", self.open_help, descr="Open the help documentation", color=ut.get_color("help") )
        ## button to update the metadata and channel and go to next step
        update_btn = fwid.add_button( "Update", self.update_metadata, descr="Update the metadata and channel choice", color=ut.get_color("go") )
        btn_line = fwid.double_button( update_btn, help_btn )

        layout.addLayout( btn_line )
        self.setLayout( layout )
        self.show_helptext()

    def update_metadata( self ):
        """ Update the metadata based on selected parameters """

        ## read the parameters
        self.mig.scaleXY = float( self.scaleXY.text() )
        self.mig.scaleZ = float( self.scaleZ.text() )
        if self.zdirection.currentText() == "top high z":
            self.mig.zdirection = -1
        else:
            self.mig.zdirection = 1
        self.mig.junchan = self.junchan.currentText()
        if self.mig.junchan == "None":
            self.mig.junchan = None
        else:
            self.mig.junchan = int(self.mig.junchan)
        self.mig.nucchan = self.nucchan.currentText()
        if self.mig.nucchan == "None":
            self.mig.nucchan = None
        else:
            self.mig.nucchan = int(self.mig.nucchan)

        ## udpate the view scale        
        for chan in range( self.mig.nbchannels ):
            self.viewer.layers['originalChannel'+str(chan)].scale = [self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY]

        ## clean the view
        ut.remove_all_widget( self.viewer )
        ut.removeOverlayText( self.viewer )

        ## update the config
        self.cfg.addGroupParameter("ImageScalings")
        self.cfg.addParameter("ImageScalings", "scalexy", self.mig.scaleXY)
        self.cfg.addParameter("ImageScalings", "scalez", self.mig.scaleZ)
        self.cfg.addParameter("ImageScalings", "direction", self.mig.zdirection)
        self.cfg.addParameter("ImageScalings", "junction_channel", self.mig.junchan)
        self.cfg.addParameter("ImageScalings", "nuclei_channel", self.mig.nucchan)
        self.viewer.grid.enabled = False
        self.cfg.write_parameterfile()

        if self.load_previous.isChecked():
            self.ffeats.load_all_previous_files()
            return None
        
        if self.mig.should_separate():
            ut.show_info("Junctions and nuclei staining in the same color channel, need to separate them")
            separation = Separation( self.ffeats ) 
            self.viewer.window.add_dock_widget(separation, name="Separate")

        if not ut.has_widget( self.viewer, "Main" ):
            self.ffeats.getChoices()

    def show_helptext( self ):
        """ Show scalings choice help text """
        help_text = ut.help_shortcut( "view" )
        help_text += ut.scale_shortcuts()
        ut.showOverlayText( self.viewer, help_text )

    def open_help( self ):
        """ Open doc webpage for Image scalings """
        ut.show_documentation_page( "Image-scalings" )

class CropImage( QWidget ):
    """ Crop the image and the associated files """
    
    def __init__( self, ffeats ):
        """ Interface to crop the image and the associated files """

        self.viewer = ffeats.viewer
        self.mig = ffeats.mig
        self.cfg = ffeats.cfg
        self.crop_layer = None

        self.add_shape_layer()

        super().__init__()
        layout = QVBoxLayout()

        ## get the name of the output crop
        line_crop, self.crop_name = fwid.value_line( "Cropped name:", self.mig.crop_name(), descr="Choose the name of the output cropped image" )
        layout.addLayout( line_crop )

        ## button to launch the crop
        crop_btn = fwid.add_button( "Do crop", self.go_crop, descr="Crop the image from the drawn rectangle and selected parameter", color=ut.get_color("go") )
        layout.addWidget( crop_btn )

        self.setLayout( layout )


    def go_crop( self ):
        """ Performs the crop """
        if self.crop_layer is None:
            print( "No crop layer" )
            return
        ## Get the rectangle to crop
        if len(self.crop_layer.selected_data) > 0:
            crop_rect = self.crop_layer.data[list(self.crop_layer.selected_data)[0]]
        elif len(self.crop_layer.data) > 0:
            crop_rect = self.crop_layer.data[0]
        else:
            ut.show_warning("No drawn rectangle, cannot crop")
            return

        ## Crop and save the main image 
        crop_rect = crop_rect / self.mig.scaleXY ## adjust to coordinates
        crop_img = self.crop_rectangle( self.mig.image, crop_rect ) 
        crop_name = self.crop_name.text()
        self.mig.save_image( crop_img, imagename=crop_name, hasZ=True, imtype="uint16" )

        ## Crop and save the cell segmentation if any
        crop_junc = None
        if (self.mig.pop is not None) and (self.mig.pop.imgcell is not None):
            crop_junc = self.crop_rectangle( self.mig.pop.imgcell, crop_rect )
        else:
            juncfile = self.mig.junction_filename(dim=2, ifexist=True)
            if os.path.exists( juncfile ):
                crop_junc, scaleX, scaleZ, names = ut.open_image( juncfile, verbose=True )

        if crop_junc is not None:
            crop_junc_name = self.get_name( "_cells2D.tif" )
            self.mig.save_image( crop_junc, imagename=crop_junc_name, hasZ=False, imtype="uint16" )

        ## Crop and save the nuclei segmentation if any
        crop_nuc = None
        if (self.mig.pop is not None) and (self.mig.pop.imgnuc is not None):
            crop_nuc = self.crop_rectangle( self.mig.pop.imgnuc, crop_rect )
        else:
            nucfile = self.mig.nuclei_filename(ifexist=True)
            if os.path.exists( nucfile ):
                crop_nuc, scaleX, scaleZ, names = ut.open_image( nucfile, verbose=True )

        if crop_nuc is not None:
            crop_nuc_name = self.get_name( "_nuclei.tif" )
            self.mig.save_image( crop_nuc, imagename=crop_nuc_name, hasZ=True, imtype="uint16" )

        ## crop other image files if they exist
        files = ["_junction_projection.tif", "_junctionsStaining.tif", "_nucleiStaining.tif" ]
        for i in range(self.mig.nbchannels):
            files = files + ["_RNA"+str(i)+".tif"]
        for cfile in files:
            filename = self.mig.build_filename( cfile )
            if os.path.exists( filename ):
                tocrop, scaleX, scaleZ, names = ut.open_image( filename, verbose=True )
                crop = self.crop_rectangle( tocrop, crop_rect )
                cropfile_name = self.get_name( cfile )
                z = len(crop.shape) > 2
                self.mig.save_image( crop, imagename=cropfile_name, hasZ=z )
                

        print("Crop RNA segmentation from csv ")
        ## Crop RNA segmentation if it exists

        for chan in range(self.mig.nbchannels):
            rnafilename = self.mig.rna_filename( chan=chan, how=".csv", ifexist=True )
            crop_spots = []
            if os.path.exists( rnafilename ):
                ## Load the RNA spots
                rnaspotDict = ut.load_dictlist(rnafilename, verbose=True)
                for rnaspot in rnaspotDict:
                    if rnaspot.get("X") is not None:
                        if int(rnaspot["X"]) >= crop_rect[0][0] and int(rnaspot["X"]) <= crop_rect[2][0] and \
                           int(rnaspot["Y"]) >= crop_rect[0][1] and int(rnaspot["Y"]) <= crop_rect[2][1]:
                            ## RNA spot in the crop rectangle, keep it and adjust coordinates
                            rnaspot["X"] = int(rnaspot["X"]) - crop_rect[0][0]
                            rnaspot["Y"] = int(rnaspot["Y"]) - crop_rect[0][1]
                            crop_spots.append( rnaspot )
                ## Save the cropped RNA spots
                if len(crop_spots) > 0:
                    crop_rna_name = self.get_name( "_RNA"+str(chan)+".csv" )
                    ut.write_dict( crop_rna_name, crop_spots ) 
                    ut.show_info( "Cropped RNA spots saved in: "+crop_rna_name )


    def get_name( self, endname ):
        """ Build the name of the output file """
        imgname, imgdir, resdir = ut.extract_names( self.crop_name.text() )
        return os.path.join( resdir, imgname+endname )

    def crop_rectangle( self, img, crop_rect ):
        """ Apply the rectangle to crop img """
        ## 4d image: channel, z, y, x
        if len(img.shape) > 3:
            return img[ :,:,int(crop_rect[0][0]):int(crop_rect[2][0]), int(crop_rect[0][1]):int(crop_rect[2][1]) ]
        ## 3d image
        if len(img.shape) > 2:
            return img[ :,int(crop_rect[0][0]):int(crop_rect[2][0]), int(crop_rect[0][1]):int(crop_rect[2][1]) ]
        ## 2d image
        return img[ int(crop_rect[0][0]):int(crop_rect[2][0]), int(crop_rect[0][1]):int(crop_rect[2][1]) ]



    def finish( self ):
        """ Finish this option, close everything """
        ut.removeOverlayText( self.viewer )
        ut.remove_widget( self.viewer, "CropImage" )
        ut.remove_layer( self.viewer, "Crop" )

    def add_shape_layer( self ):
        """ Add the layer to draw the rectangle for the crop """
        self.viewer.dims.ndisplay = 2 ## force 2D view for shape drawing
        self.crop_layer = self.viewer.add_shapes( [], name="Crop" )
        text = "Draw rectangle to crop"
        self.crop_layer.mode = "add_rectangle"
        ut.showOverlayText( self.viewer, text )


class Association( QWidget ):
    """ Interface to associate cell and nuclei together """

    def __init__( self, viewer, mig, cfg ):
        """ Creates the interface """
        super().__init__()
        self.viewer = viewer
        self.mig = mig
        self.cfg = cfg
    
        # initialize parameters, show help msg
        text = "Find the nucleus associated with each apical cell \n"
        print("******* Associate apical cells and nuclei together ******")
        ut.remove_widget( self.viewer, "Associating")
        ut.showOverlayText(self.viewer, text)
        ## load parameters
        defmethod = "Calculate association"
        distasso = 30.0
        assojuncfile = self.mig.junction_filename(dim=2, ifexist=True)
        assonucfile = self.mig.nuclei_filename(ifexist=True)
        paras = self.cfg.read_parameter_set("Association")
        if paras is not None:
            if "method" in paras:
                defmethod = paras["method"]
            if "distance_toassociate_micron" in paras:
                distasso = float(paras["distance_toassociate_micron"])
            if "associated_junctions" in paras:
                assojuncfile = paras["associated_junctions"]
            if "associated_nuclei" in paras:
                assonucfile = paras["associated_nuclei"]
        if os.path.exists(assojuncfile):
            defmethod = "Load association"

        layout = QVBoxLayout()
        ## choice of association method
        line_method, self.method = fwid.list_line( "Association method:", descr="Choose the method to associate cells and nuclei" )
        self.method.addItems( ["Load association", "Calculate association"] )
        layout.addLayout( line_method )
        self.method.currentTextChanged.connect( self.update_visibility )
        self.method.setCurrentText(defmethod)

        ## Loading parameters
        self.load_group, load_layout = fwid.group_layout( "Loading parameters", descr="Parameters to choose files to load" )
        ## choice of cells file
        cell_line, self.cell_file = fwid.file_line( "Associated cell file:", assojuncfile, "Choose cell file", descr="Choose the file containing the associated cells" )
        load_layout.addLayout( cell_line )
        ## choice of nuclei file
        nuclei_line, self.nuclei_file = fwid.file_line( "Associated nuclei file:", assonucfile, "Choose nuclei file", descr="Choose the file containing the associated nuclei" )
        load_layout.addLayout( nuclei_line )
        self.load_group.setLayout( load_layout )
        layout.addWidget( self.load_group 
                         )
        ## Calculation parameters
        self.calc_group, calc_layout = fwid.group_layout( "Calculate association parameters", descr="Parameters for the association calculation" )
        ## max distance for association
        dist_line, self.max_distance = fwid.value_line( "Max association distance (um):", distasso, descr="Set the maximum distance between nucleus and cell for association" )
        calc_layout.addLayout( dist_line )
        self.calc_group.setLayout( calc_layout )
        layout.addWidget( self.calc_group )

        ## btn go assocation
        btn_go = fwid.add_button( "Go association", self.go_association, descr="Associate the cells and nuclei based on the selected method", color=ut.get_color("go") )
        ## help button
        btn_help = fwid.add_button( "Help", self.open_help, descr="Open the help documentation", color=ut.get_color("help") )
        ##line with the buttons
        btn_line = fwid.double_button( btn_go, btn_help )

        layout.addLayout( btn_line )
        self.setLayout( layout )
        self.update_visibility()

    def update_visibility( self ):
        """ Update the visibility of parameters based on method """
        booly = self.method.currentText() == "Load association"
        self.load_group.setVisible( booly )
        self.calc_group.setVisible( not booly )
    
    def open_help(self):
        """ Open the Wiki documentation page """
        ut.show_documentation_page("Associate")
    
    def load_association( self ):
        """ Load association from files """
        self.mig.load_segmentation(self.cell_file.text())
        self.mig.popFromJunctions()
        self.mig.load_segmentation_nuclei(self.nuclei_file.text())
        self.mig.popNucleiFromMask()
        ut.remove_widget(self.viewer, "Associating")
        ut.removeOverlayText(self.viewer)
        self.end_association()

    def go_association( self ):
        """ Perform association with selected paramters """

        if not self.mig.hasCells():
            ut.show_error("No junctions were segmented/loaded. Do it before")
            return
        if not self.mig.hasNuclei():
            ut.show_error("No nuclei were segmented/loaded. Do it before")
            return

        ## save current parameter to conf file
        self.cfg.addGroupParameter("Association")
        self.cfg.addParameter("Association", "method", self.method.currentText())
        self.cfg.addParameter("Association", "associated_junctions", self.cell_file.text())
        self.cfg.addParameter("Association", "associated_nuclei", self.nuclei_file.text())
        self.cfg.addParameter("Association", "distance_toassociate_micron", self.max_distance.text())
        self.cfg.write_parameterfile()

        ## perform selected method
        if self.method.currentText() == "Calculate association":
            start_time = ut.get_time()
            ut.showOverlayText(self.viewer, "Doing junction-nuclei association...")
            ut.show_info("Associate "+str(self.mig.nbCells())+" junctions with nuclei...")
            pbar = ut.start_progress( self.viewer, total=2, descr="Calculating association..." )
            self.calc_association(distance=float(self.max_distance.text()), pbar=pbar)
            ut.close_progress( self.viewer, pbar )
            ut.show_duration( start_time, "Association calculated in ")
    
        else:
            ut.show_info("Load association from files")
            self.load_association()

    def calc_association( self, distance, pbar=None ):
        self.mig.go_association(distance=distance, pbar=pbar)
        ut.remove_widget(self.viewer, "Associating")
        self.end_association()

    def end_association(self):
        """ Automatic association finished, go to manual correction step """
    
        ut.show_info("Correct association if necessary")
        ut.removeOverlayText( self.viewer )
        ut.remove_layer(self.viewer, "CellContours")
        ut.remove_layer(self.viewer, "CellNuclei")
        self.viewer.add_labels( self.mig.getJunctionsImage3D(), name="CellContours", blending="additive", scale=(self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY) )
        self.viewer.layers["CellContours"].editable = False
        self.viewer.add_labels( self.mig.nucmask, name="CellNuclei", blending="additive", scale=(self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY) )
        self.viewer.layers["CellNuclei"].n_edit_dimensions = 3
        
        corr = AssociateCN( self.viewer, self.mig, self.cfg, self.viewer.layers["CellContours"], self.viewer.layers["CellNuclei"] )
        self.viewer.window.add_dock_widget( corr, name="Edit association" )

class AssociateCN( QWidget ):
    """ Interface to correct the cell-nucleus association """

    def __init__( self, viewer, mig, cfg, layerJunc, layerNuc, shapeName="JunctionNames", dim=3 ):
        """ Interface to correct the cell-nucleus association """
        super().__init__()
        self.viewer = viewer
        self.mig = mig
        self.cfg = cfg
        self.layerJun = layerJunc
        self.layerNuc = layerNuc
        self.shapeName = shapeName
        self.dim = dim

        layout = QVBoxLayout()
        ## choice of nucleus
        nc_line, self.nucleus, self.cell = fwid.double_value_line( "Associate nucleus: ", 0, "with cell: ", 0, descr="Choose the nucleus label and cell label to associate together" )
        layout.addLayout(nc_line)
        ## button go
        btn_go = fwid.add_button( "Associate now", self.associate, descr="Associate the current nucleus and cell together" )
        ## help button
        btn_help = fwid.add_button( "Help", self.show_help, descr="Show help for cell-nucleus association", color=ut.get_color("help") )
        line = fwid.double_button( btn_go, btn_help )
        layout.addLayout(line)

        ## show cell names
        self.show_cellnames = fwid.add_check( "Show cell names", False, self.show_names, descr="Show/Hide the cell names on the junction layer" )
        ## synchronize the layers
        self.resync_layers = fwid.add_check( "Synchronize layers", False, self.update_synchronize_layers, descr="Synchronize the two Cell and Nuclei layer displays" )
        check_line = fwid.double_widget( self.show_cellnames, self.resync_layers )
        layout.addLayout(check_line)

        ## Save and done buttons
        self.save = fwid.add_button( "Save association", self.save_association, descr="Save the current association results to files", color=ut.get_color("save") )
        self.done = fwid.add_button( "Association done", self.association_done, descr="Finish the association step and remove the associated layers", color=ut.get_color("done") )
        btn_line = fwid.double_button( self.save, self.done )
        layout.addLayout( btn_line )
        self.setLayout(layout)

        self.show_message()
        self.activate_bindings()

    def save_association( self ):
        """ Save the current results to files """
        nuc_filename = self.mig.nuclei_filename( ifexist=False)
        self.mig.save_image( self.mig.nucmask, nuc_filename, hasZ=True )
        self.mig.popNucleiFromMask( associate=True )
        self.mig.save_results()

    def association_done( self ):
        """ Finish this step """
        junc3D = self.viewer.layers["CellContours"].data
        self.mig.junmask = np.max(junc3D, axis=0)
        ut.remove_layer( self.viewer,"CellContours" )
        ut.remove_layer( self.viewer,"CellNuclei")
        ut.remove_layer( self.viewer,"JunctionNames")
        ut.remove_widget( self.viewer, "Edit association")
        ut.removeOverlayText( self.viewer)


    def show_message( self ):
        """ show update messages """
        help_text = "<Control+Left-click> to select a nucleus value \n"
        help_text = help_text + "<Right-click> to choose the cell to associate with \n"
        help_text = help_text + "<c> to apply current association \n"
        help_text = help_text + "<l> to show/hide cell labels \n"
        help_text = help_text + "<s> to synchronize junctions and nuclei view \n"
        help_text = help_text + "<u> to unsynchronize junctions and nuclei view \n"
        help_text += "  <Ctrl-c>/<Ctrl-d> increase/decrease NUCLEI label contour \n"
        help_text += "  <Shift-c>/<Shift-d> increase/decrease JUNCTIONS label contour \n"
        header = ut.helpHeader( self.viewer, "CellNuclei" )
        ut.showOverlayText(self.viewer, header+help_text)
        print("\n ---- Association editing ---- ")

    def associate( self ):
        """ Associate current nucleus and cell together """
        nucleus = int(self.nucleus.text())
        cell = int(self.cell.text())
        print("Associate nucleus "+str(nucleus)+" with cell "+str(cell))
        self.mig.associateCN(int(nucleus), int(cell))
        self.viewer.layers["CellNuclei"].refresh()

    def show_help( self ):
        """ Open help for manual correction step """
        ut.show_documentation_page("Associate#Manual-correction")

    def activate_bindings( self ):
        """ Activate specific shortcuts for this step """
    
        @self.layerNuc.bind_key('c', overwrite=True)
        def associateBis(layer):
            nucleus = int(self.nucleus.text())
            cell = int(self.cell.text())
            if nucleus == 0 or cell == 0:
                print("One value is zero, ignore association")
                return
            print("Associate nucleus "+str(nucleus)+" with cell "+str(cell))
            self.mig.associateCN(nucleus, cell)
            self.viewer.layers["CellNuclei"].refresh()
    
        @self.layerNuc.bind_key('l', overwrite=True)
        @self.layerJun.bind_key('l', overwrite=True)
        def showCellNames(layer):
            self.show_cell_names()
    
        @self.layerNuc.bind_key('s', overwrite=True)
        @self.layerJun.bind_key('s', overwrite=True)
        def synchronizeLayers(layer):
            self.synchronize_layers()
    
        @self.layerNuc.bind_key('u', overwrite=True)
        @self.layerJun.bind_key('u', overwrite=True)
        def unsynchronizeLayers(layer):
            self.unsynchronize_layers()

        @self.layerNuc.bind_key('Control-c', overwrite=True)
        @self.layerJun.bind_key('Control-c', overwrite=True)
        def contour_increase(layer):
            if self.layerNuc is not None:
                self.layerNuc.contour = self.layerNuc.contour + 1

        @self.layerNuc.bind_key('Shift-c', overwrite=True)
        @self.layerJun.bind_key('Shift-c', overwrite=True)
        def contour_decrease(layer):
            if self.layerJun is not None:
                if self.layerJun.contour > 0:
                    self.layerJun.contour = self.layerJun.contour - 1

        @self.layerNuc.bind_key('Shift-d', overwrite=True)
        @self.layerJun.bind_key('Shift-d', overwrite=True)
        def contour_decrease(layer):
            if self.layerJun is not None:
                if self.layerJun.contour > 0:
                    self.layerJun.contour = self.layerJun.contour - 1

        @self.layerNuc.bind_key('Control-d', overwrite=True)
        @self.layerJun.bind_key('Control-d', overwrite=True)
        def contour_decrease(layer):
            if self.layerNuc is not None:
                if self.layerNuc.contour > 0:
                    self.layerNuc.contour = self.layerNuc.contour - 1

        # Handle click or drag events separately
        @self.layerNuc.mouse_drag_callbacks.append
        def click(layer, event):
            if event.type == "mouse_press":
                if event.button == 2:
                    # right click
                    value = self.layerJun.get_value(position=event.position, view_direction = event.view_direction, dims_displayed=event.dims_displayed, world=True)
                    self.cell.setText( str(value) )
                if (event.button == 1) and ("Control" in event.modifiers): 
                    ## associate nucleus with cell
                    value = self.layerNuc.get_value(position=event.position, view_direction = event.view_direction, dims_displayed=event.dims_displayed, world=True)
                    self.nucleus.setText( str(value) )
                    self.layerNuc.selected_label = value

    def show_names( self ):
        """ Show/Hide cell names """
        if self.show_cellnames.isChecked():
            if not self.shapeName in self.viewer.layers:
                self.show_cell_names()
        else:
            ut.remove_layer( self.viewer, self.shapeName )

    def update_synchronize_layers( self ):
        """ synchronize on/off of the two layers """
        if self.resync_layers.isChecked():
            self.synchronize_layers() 
        else:
            self.unsynchronize_layers()

    def show_cell_names( self ):
        """ Show cell names inside each cell """
        if self.shapeName in self.viewer.layers:
            ut.remove_layer(self.viewer, self.shapeName)
            self.show_cellnames.setChecked( False )
        else:
            ut.get_bblayer(self.layerJun, self.shapeName, self.dim, self.viewer, self.mig)
            self.show_cellnames.setChecked( True )

    def synchronize_layers( self ):
        """ Synchronize the two layers """
        self.viewer.layers.link_layers((self.layerJun, self.layerNuc), ('selected_label', 'n_edit_dimensions', 'visible', 'refresh', 'contiguous'))
        self.layerJun.show_selected_label = True
        self.layerNuc.show_selected_label = True
        self.resync_layers.setChecked( True )

    def unsynchronize_layers(self):
        """ Unsynchronize the two layers """
        self.viewer.layers.unlink_layers()
        self.layerJun.show_selected_label = False
        self.layerNuc.show_selected_label = False
        self.resync_layers.setChecked( False )

class Separation( QWidget ):
    """ Interface to separate junctions and nuclei if they are in the same channel """

    def __init__( self, ffeats ):
        """ Separation GUI """
        super().__init__()

        self.viewer = ffeats.viewer
        self.mig = ffeats.mig
        self.cfg = ffeats.cfg

        self.show_text()
        ## load parameters
        paras = {}
        paras["tophat_radxy"] = 4
        paras["tophat_radz"] = 1
        paras["outlier_thres"] = 40
        paras["smooth_nucleixy"] = 2
        paras["smooth_nucleiz"] = 2
        paras["sepanet_path"] = os.path.join(".", "sepaNet")
    
        load_paras = self.cfg.read_parameter_set("Separate")
        if load_paras is not None:
            if "method" in load_paras:
                paras["method"] = load_paras["method"]
            if "tophat_radxy" in load_paras:
                paras["tophat_radxy"] = int(load_paras["tophat_radxy"])
            if "tophat_radz" in load_paras:
                paras["tophat_radz"] = int(load_paras["tophat_radz"])
            if "outlier_thres" in load_paras:
                paras["outlier_thres"] = int(load_paras["outlier_thres"])
            if "smooth_nucleixy" in load_paras:
                paras["smooth_nucleixy"] = int(load_paras["smooth_nucleixy"])
            if "smooth_nucleiz" in load_paras:
                paras["smooth_nucleiz"] = int(load_paras["smooth_nucleiz"])
            if "sepanet_path" in load_paras:
                paras["sepanet_path"] = load_paras["sepanet_path"]
    
        defmethod = "SepaNet"
        if "method" in paras:
            defmethod = paras["method"]
        separated_junctionsfile = self.mig.separated_junctions_filename(ifexist=True)
        if separated_junctionsfile != "":
            defmethod = "Load"

        ## create interface
        layout = QVBoxLayout()
        ## choose method for separation
        method_line, self.method = fwid.list_line( "Separation method:", descr="Choose the method to separate junctions and nuclei" )
        self.method.addItems( ["Load", "Tophat filter", "SepaNet"] )
        layout.addLayout( method_line )

        ## parameters for loading
        self.load_group, load_layout = fwid.group_layout( "Loading parameters", descr="Parameters to choose files to load" )
        filejunc_line, self.separated_juncfile = fwid.file_line( "Separated junction file:", separated_junctionsfile, "Choose separated junction file", descr="Choose the file containing the separated junctions" )
        filenuc_line, self.separated_nucfile = fwid.file_line( "Separated nuclei file:", self.mig.separated_nuclei_filename(ifexist=True), "Choose separated nuclei file", descr="Choose the file containing the separated nuclei" )
        load_layout.addLayout( filejunc_line )
        load_layout.addLayout( filenuc_line )
        self.load_group.setLayout( load_layout )
        layout.addWidget( self.load_group )

        ## parameters for SepaNet
        self.sepanet_group, sepanet_layout = fwid.group_layout( "SepaNet parameters", descr="Parameters for SepaNet neural network separation" )
        sepamodel_line, self.sepanet_modelpath = fwid.dir_line( "SepaNet model path:", os.path.join(paras["sepanet_path"]), "Choose SepaNet model path", descr="Choose the path to the SepaNet model" )
        sepanet_layout.addLayout( sepamodel_line )
        self.sepanet_group.setLayout( sepanet_layout )
        layout.addWidget( self.sepanet_group )

        ## parameters for top hat filter
        self.tophat_group, tophat_layout = fwid.group_layout( "Top hat filter parameters", descr="Parameters for top hat filter separation" )
        radxy_line, self.tophat_radxy = fwid.value_line( "Tophat radius XY (pixels):", paras["tophat_radxy"], descr="Set the radius in XY for the tophat filter" )
        radz_line, self.tophat_radz = fwid.value_line( "Tophat radius Z (pixels):", paras["tophat_radz"], descr="Set the radius in Z for the tophat filter" )
        tophat_layout.addLayout( radxy_line )
        tophat_layout.addLayout( radz_line )
        #" outlier para"
        outlier_line, self.outlier_thres = fwid.value_line( "Outlier threshold:", paras["outlier_thres"], descr="Set the outlier threshold for the tophat filter" )
        tophat_layout.addLayout( outlier_line )
        # smoothing
        smoothxy_line, self.smooth_nucleixy = fwid.value_line( "Nuclei smoothing XY (pixels):", paras["smooth_nucleixy"], descr="Set the smoothing in XY for the nuclei after tophat filter" )
        smoothz_line, self.smooth_nucleiz = fwid.value_line( "Nuclei smoothing Z (pixels):", paras["smooth_nucleiz"], descr="Set the smoothing in Z for the nuclei after tophat filter" )
        tophat_layout.addLayout( smoothxy_line )
        tophat_layout.addLayout( smoothz_line )
        self.tophat_group.setLayout( tophat_layout )
        layout.addWidget( self.tophat_group )

        ## Go separate buton
        separate_btn = fwid.add_button( "Go separation", self.separate_go, descr="Separate the junctions and nuclei based on the selected method", color=ut.get_color("go") )
        layout.addWidget( separate_btn )

        ## save results
        save_btn = fwid.add_button( "Save separated", self.save_separated_staining, descr="Save the separated junctions and nuclei to files", color=ut.get_color("save") )
        help_btn = fwid.add_button( "Help", self.separate_help, descr="Open the help documentation", color=ut.get_color("help") )
        btn_line = fwid.double_button( save_btn, help_btn )
        layout.addLayout( btn_line )

        ## close layer when finishing
        self.close_layers = fwid.add_check( "Close layers", True, None, descr="Close the created junction and nuclei layers at the end of this step" )
        btn_done = fwid.add_button( "Separation done", self.separate_done, descr="Finish the separation step and remove the created layers", color=ut.get_color("done") )
        done_line = fwid.double_widget( self.close_layers, btn_done )
        layout.addLayout( done_line )

        self.setLayout( layout )
        self.method.currentTextChanged.connect( self.update_visibility )
        self.method.setCurrentText( defmethod )
        self.update_visibility()

    def update_visibility( self ):
        """ update visibility of parameters option """
        self.load_group.setVisible( self.method.currentText() == "Load" )
        self.sepanet_group.setVisible( self.method.currentText() == "SepaNet" )
        self.tophat_group.setVisible( self.method.currentText() == "Tophat filter" )

    def show_text( self ):
        """ Show help message """
        text = "Separate the junction and nuclei staining that are in the same channel \n"
        text += "Creates two new layers, junctionStaining and nucleiStaining \n"
        text += "Tophat filter option separate the signals based on morphological filtering \n"
        text += "Check the \'close layers\' box if you want the two created layers to be closed at the end of this step \n"
        text += "SepaNet option separate the signals with trained neural networks \n"
        ut.showOverlayText( self.viewer, text )
        ut.show_info("********** Separating the junction and nuclei staining ***********")

    def separate_help( self ):
        """ Open the documentation page """
        ut.show_documentation_page("Separate-junctions-and-nuclei")
    
    def save_separated_staining( self ):
        """ Save the two result images """
        if "junctionsStaining" in self.viewer.layers:
            outname = self.mig.separated_junctions_filename()
            self.mig.save_image( self.viewer.layers["junctionsStaining"].data, outname, hasZ=True, imtype="uint8" )
        if "nucleiStaining" in self.viewer.layers:
            outname = self.mig.separated_nuclei_filename()
            self.mig.save_image( self.viewer.layers["nucleiStaining"].data, outname, hasZ=True, imtype="uint8" )

    def separate_done( self ):
        """ Finish the step """
        ut.remove_widget( self.viewer, "Separate" )
        self.cfg.addGroupParameter("Separate")
        self.cfg.addParameter("Separate", "method", self.method.currentText() )
        self.cfg.addParameter("Separate", "sepanet_path", self.sepanet_modelpath.text() )
        self.cfg.addParameter("Separate", "tophat_radxy", int(self.tophat_radxy.text()))
        self.cfg.addParameter("Separate", "tophat_radz", int(self.tophat_radz.text()))
        self.cfg.addParameter("Separate", "outlier_thres", int(self.outlier_thres.text()))
        self.cfg.addParameter("Separate", "smooth_nucleixy", int(self.smooth_nucleixy.text()))
        self.cfg.addParameter("Separate", "smooth_nucleiz", int(self.smooth_nucleiz.text()))
        self.cfg.write_parameterfile()
        if self.close_layers.isChecked():
            ut.remove_layer(self.viewer, "junctionsStaining")
            ut.remove_layer(self.viewer, "nucleiStaining")
        return 1
    
    def separate_go( self ):
        """ Perform separation with selected method and parameters """
        ut.showOverlayText( self.viewer, "Discriminating between nuclei and junctions...")
        ut.hide_color_layers(self.viewer, self.mig)
        ut.remove_layer(self.viewer,"junctionsStaining")
        ut.remove_layer(self.viewer,"nucleiStaining")
        if self.method.currentText() == "Tophat filter":
            self.discriminating()
        if self.method.currentText() == "SepaNet":
            self.sepaneting( self.sepanet_modelpath.text() )
        if self.method.currentText() == "Load":
            self.load_separated( self.separated_juncfile.text(), self.separated_nucfile.text(), end_dis=True)

    def load_separated( self, junfile, nucfile, end_dis=True ):
        """ Load the two separated staining from files """
        self.mig.load_separated_staining( junfile, nucfile )
        ut.show_info("Separated stainings loaded")
        if end_dis:
            self.end_discrimination()

    def sepaneting( self, sepanet_dir ):
        """ Separate the junction and nuclei stainign with SepaNet trained networks """
        self.viewer.window._status_bar._toggle_activity_dock(True)
        self.mig.separate_with_sepanet( sepanet_dir )
        self.viewer.window._status_bar._toggle_activity_dock(False)
        self.end_discrimination()

    def discriminating( self ):
        """ do top hat discrimination """
        self.mig.separate_junctions_nuclei( wth_radxy=int(self.tophat_radxy.text()),
                wth_radz = int(self.tophat_radz.text()),
                rmoutlier_threshold=int(self.outlier_thres.text()),
                smoothnucxy=int(self.smooth_nucleixy.text()),
                smoothnucz=int(self.smooth_nucleiz.text()) )
        self.end_discrimination()

    def end_discrimination(self):
        """ Show resulting separated stainings """
        self.viewer.add_image( self.mig.junstain, name="junctionsStaining", blending="additive", scale=(self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY), colormap="red" )
        self.viewer.add_image( self.mig.nucstain, name="nucleiStaining", blending="additive", scale=(self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY), colormap="blue" )
        ut.removeOverlayText(self.viewer)

class CytoplasmMeasure( QWidget ):
    """
    Measure the cytoplasm intensity of the cells
    """

    def __init__( self, viewer, mig, cfg ):
        """ GUI to measure cyto intensity """
        super().__init__()
        self.viewer = viewer
        self.mig = mig
        self.cfg = cfg
        import ast

        ## show message
        text = "Measure cytoplasmic intensity close to the apical surface \n"
        text += "Choose the channel to measure in the \'cyto_channels\' parameter \n"
        text += "z_thickness is the number of z slices below the apical surface used for the measure \n "
        text += "Use the rectangle to estimate background intensity. The value will be averaged from the z_thickness slices below the rectangle \n"
        ut.showOverlayText( self.viewer, text )
        print("********** Measure cytoplasmic intensities **************")

        for layer in self.viewer.layers:
            layer.visible = False

        # load parameters
        self.meanz = self.mig.getAverageCellZ()
        paras = {}
        paras["cytoplasmic_channels"] = [self.mig.free_channel()]
        paras["save_measures_table"] = True
        paras["show_measures_image"] = True
        paras["z_thickness"] = 3
        load_paras = self.cfg.read_parameter_set("MeasureCytoplasmic")
        if load_paras is not None:
            if "z_thickness" in load_paras:
                paras["z_thickness"] = int(load_paras["z_thickness"])
            if "cytoplasmic_channels" in load_paras:
                paras["cytoplasmic_channels"] = ast.literal_eval( load_paras["cytoplasmic_channels"].strip() )

        ## create the GUI
        layout = QVBoxLayout()
        ## choose channel(s) to measure
        chan_line, self.channels = fwid.add_multiple_list( "Channel(s ):", descr="Choose the channel(s) to measure cytoplasmic intensity")
        self.channels.addItems( [str(chan) for chan in range(self.mig.nbchannels)] )
        self.channels.itemSelectionChanged.connect( self.update_channel_selection )
        layout.addLayout( chan_line )

        ## choose thickness of measure
        thick_line, self.thickness = fwid.value_line( "Z thickness (slices):", paras["z_thickness"], descr="Set the number of z slices below the apical surface used for the measure" )
        layout.addLayout( thick_line )

        ## btn go measure
        measure_btn = fwid.add_button(" Measure cytoplasmic intensity", self.measure_cyto, descr="Measure intensity in selected channels" )
        layout.addWidget(measure_btn)

        ## btn help, done
        help_btn = fwid.add_button(" Help", self.show_cytomeas_doc, descr="Show help for cytoplasmic intensity measurement", color=ut.get_color("help") )
        done_btn = fwid.add_button(" Measurement done", self.measure_done, descr="Finish the cytoplasmic intensity measurement step", color=ut.get_color("done") )
        btn_line = fwid.double_button( help_btn, done_btn )
        layout.addLayout(btn_line)

        self.setLayout(layout)

    def measure_cyto( self ):
        """ Launch the measurement and show the resulting maps. Save the results """
        channels = self.channels.selectedItems()
        channels = [int(chan.text()) for chan in channels]
        bgrois = []
        for chan in channels:
            layer = self.viewer.layers["backgroundRectangle_"+str(chan)]
            bgrois.append(layer.data)
            layer.visible = False
        
        self.cfg.addGroupParameter("MeasureCytoplasmic")
        self.cfg.addParameter("MeasureCytoplasmic", "cytoplasmic_channels", list(channels))
        self.cfg.addParameter("MeasureCytoplasmic", "z_thickness", int(self.thickness.text()))
        self.cfg.write_parameterfile()
        ut.removeOverlayText(self.viewer)

        results = self.mig.measureCytoplasmic(channels, bgrois, int(self.thickness.text()))
        self.mig.save_results()
        self.show_measure_images( results )

    def show_measure_images( self, results ):
        """ Draw and show the measurement maps """
        channels = self.channels.selectedItems()
        channels = [int(chan.text()) for chan in channels]
        for i, chan in enumerate(channels):
            if "Intensity"+str(chan) in self.viewer.layers:
                ut.remove_layer(self.viewer, "CytoplasmicNormalisedIntensity"+str(chan))
            cytomes = self.mig.drawCytoplasmicMeasure( chan, results )
            cproj = self.viewer.add_image(cytomes, name="CytoplasmicNormalisedIntensity"+str(chan), scale=(self.mig.scaleXY, self.mig.scaleXY), colormap=ut.colormapname(chan), blending="additive")
            cproj.contrast_limits=ut.quantiles(cytomes)

    def update_channel_selection( self ):
        """ the selected channels has changed, update the background rectangles """
        channels = self.channels.selectedItems()
        channels = [int(chan.text()) for chan in channels]
        dep = 20
        size = 50
        step = 30
        #QApplication.instance().processEvents()
        for layer in self.viewer.layers:
            layer.visible = False
        for pchan in range(self.mig.nbchannels):
            if pchan not in channels:
                ut.remove_layer(self.viewer, "backgroundRectangle_"+str(pchan))
        for chan in channels:
            polygon = np.array([[self.meanz, dep, dep], [self.meanz, dep, dep+size], [self.meanz, dep+size, dep+size], [self.meanz, dep+size, dep]])
            colname = ut.colormapname(chan)
            if not isinstance(colname, str):
                colname = colname.map(np.array([0.99]))
            try:
                #QApplication.instance().processEvents()
                if not "backgroundRectangle_"+str(chan) in self.viewer.layers:
                    self.viewer.add_shapes( polygon, name="backgroundRectangle_"+str(chan), ndim=3, shape_type='rectangle', edge_width=0, face_color=colname, scale=(self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY) )
                ut.show_layer(self.viewer, chan)
            except:
                ut.show_error( "Error while adding shape layer "+str(chan)+" \nPlease retry" )
                for pchan in range(self.mig.nbchannels):
                    ut.remove_layer(self.viewer, "backgroundRectangle_"+str(pchan))
                return
            dep += step
        self.viewer.dims.ndisplay = 2
        self.viewer.dims.set_point(0,self.meanz*self.mig.scaleZ)

    def measure_done(self):
        """ Finish the step """
        ut.removeOverlayText(self.viewer)
        ut.remove_widget(self.viewer, "Measure cytos")
        for chan in range(self.mig.nbchannels):
            ut.remove_layer(self.viewer, "backgroundRectangle_"+str(chan))
            ut.remove_layer(self.viewer, "CytoplasmicNormalisedIntensity"+str(chan))

    def show_cytomeas_doc(self):
        """ Open the wiki page on cytoplasmic measures """
        ut.show_documentation_page("Measure-cytoplasmic-staining")


from qtpy.QtWidgets import QPushButton, QHBoxLayout, QVBoxLayout, QWidget, QGroupBox, QCheckBox, QSpinBox, QSlider, QDoubleSpinBox, QLabel, QComboBox, QLineEdit, QFileDialog, QTabWidget, QListWidget, QAbstractItemView
import fish_feats.Utils as ut
import os, pathlib
import numpy as np
import napari
from napari.utils.translations import trans
import time
from math import ceil

class FishGrid(QWidget):
    """ Show/Load a grid to have a repere in space """

    def __init__(self, napari_viewer, mig):
        super().__init__()
        self.viewer = napari_viewer
        self.mig = mig
        
        ## define interface
        layout = QVBoxLayout()
        self.get_parameters()
        layout.addWidget(self.group_grid)
        self.setLayout(layout)

    def get_parameters(self):
        """ Get grid parameters """
        self.group_grid = QGroupBox("Grid setup")
        grid_layout = QVBoxLayout()
        ## nrows
        rows_line = QHBoxLayout()
        rows_lab = QLabel()
        rows_lab.setText("Nb rows:")
        rows_line.addWidget(rows_lab)
        self.nrows = QLineEdit()
        self.nrows.setText("4")
        rows_line.addWidget(self.nrows)
        grid_layout.addLayout(rows_line)
        ## ncols
        cols_line = QHBoxLayout()
        cols_lab = QLabel()
        cols_lab.setText("Nb columns:")
        cols_line.addWidget(cols_lab)
        self.ncols = QLineEdit()
        self.ncols.setText("4")
        cols_line.addWidget(self.ncols)
        grid_layout.addLayout(cols_line)
        ## go for grid
        btn_add_grid = QPushButton("Add grid", parent=self)
        grid_layout.addWidget(btn_add_grid)
        btn_add_grid.clicked.connect(self.add_grid)
        self.group_grid.setLayout(grid_layout)

    def add_grid(self):
        """ Create/Load a new grid and add it """
        ut.remove_layer(self.viewer, "FishGrid")
        imshape = self.mig.image_2dshape()
        if imshape is None:
            ut.show_error("Load the image first")
            return
        nrows = int(self.nrows.text())
        ncols = int(self.ncols.text())
        wid = ceil(imshape[0]/nrows)
        hei = ceil(imshape[1]/ncols)
        rects = []
        rects_names = []
        for x in range(nrows):
            for y in range(ncols):
                rect = np.array([[x*wid, y*hei], [(x+1)*wid, (y+1)*hei]])
                rects.append(rect)
                rects_names.append(chr(65+x)+"_"+str(y))
        self.viewer.add_shapes(rects, name="FishGrid", text=rects_names, face_color=[1,0,0,0], edge_color=[0.7,0.7,0.7,0.7], edge_width=3, scale=(self.mig.scaleXY, self.mig.scaleXY))




"""Segments detected regions in a chunked dask array.

Uses overlapping chunks during segmentation, and determines how to link segments
between neighboring chunks by examining the overlapping border regions.

Heavily based on dask_image.ndmeasure.label, which uses non-overlapping blocks
with a structuring element that links segments at the chunk boundaries.

Modified from: https://github.com/MouseLand/cellpose/blob/main/cellpose/contrib/distributed_segmentation.py (cellpose version 2)
"""
import functools
import operator

import dask
import dask.array as da
import numpy as np

class DistSegError(Exception):
    """Error in image segmentation."""

try:
    from dask_image.ndmeasure._utils import _label
    from sklearn import metrics as sk_metrics
except ModuleNotFoundError as e:
    raise DistSegError("Install 'cellpose[distributed]' for distributed segmentation dependencies") from e


def dask_segment(
    image,
    channels,
    diameter,
    fast_mode=False,
    use_anisotropy=True,
    iou_depth=2,
    iou_threshold=0.7,
    chunksize = (1000,1000,1000,1),    ## add the chunksize parameter for the tests
    dim = 4, ## add the dimension parameter to do 2D images as well
    do_3D = True,
    stitch_threshold=0.25,
    resample = True,
    cellprob_threshold = 0,
    flow_threshold = 0.4,
    model = None, ## Added: cellpose model
    z_axis = None, ## which axis if 3D
    tmp_file=None, ## temporary directory where to store the computed chunks: Thanks Marvin for the idea !
):
    """Use cellpose to segment nuclei in fluorescence data.

    Parameters
    ----------
    image : array of shape (z, y, x, channel)
        Image used for detection of objects
    channels : array of int with size 2
        See cellpose
    model_type : str
        "cyto" or "nuclei"
    diameter : tuple of size 3
        Approximate diameter (in pixels) of a segmented region, i.e. cell width
    fast_mode : bool
        In fast mode, network averaging, tiling, and augmentation are turned off.
    use_anisotropy : bool
        If true, use anisotropy parameter of cellpose
    iou_depth: dask depth parameter
        Number of pixels of overlap to use in intersection-over-union calculation when
        linking segments across neighboring, overlapping dask chunk regions.
    iou_threshold: float
        Minimum intersection-over-union in neighboring, overlapping dask chunk regions
        to be considered the same segment.  The region for calculating IOU is given by the
        iou_depth parameter.

    Returns:
        segments : array of int32 with same shape as input
            Each segmented cell is assigned a number and all its pixels contain that value (0 is background)
    """
    #assert image.shape[-1] in {1, 2}, image.shape
    #assert diameter[1] == diameter[2], diameter

    diameter_yx = diameter[1]
    if dim == 4:
        anisotropy = diameter[0] / diameter[1] if use_anisotropy else None
    else:
        anisotropy = 1

    image = da.asarray(image, chunks=chunksize)  ## added the chunks parameter
    
    image = image.rechunk({-1: -1})  # color channel is chunked together

    ## handle case where the overlap is larger than the chunksize
    over = iou_depth 
    if image.chunksize[0] < diameter[0]:
        depth = (over,) + (np.ceil(diameter[1]).astype(np.int64),) + (np.ceil(diameter[1]).astype(np.int64),)
    elif image.chunksize[1] < diameter[1]:
        depth = (np.ceil(diameter[1]).astype(np.int64),) + (over,) + (np.ceil(diameter[1]).astype(np.int64),)
    elif image.chunksize[2] < diameter[1]:
        depth = tuple(np.ceil(diameter).astype(np.int64)) + (over,)
    else:
        depth = tuple(np.ceil(diameter).astype(np.int64)) + (np.ceil(diameter[1]).astype(np.int64),)
    boundary = "reflect"

    # No chunking in channel direction
    image = da.overlap.overlap(image, depth + (0,), boundary)

    block_iter = zip(
        np.ndindex(*image.numblocks),
        map(
            functools.partial(operator.getitem, image),
            da.core.slices_from_chunks(image.chunks),
        ),
    )

    labeled_blocks = np.empty(image.numblocks[:-1], dtype=object)
    total = None
    for index, input_block in block_iter:
        labeled_block, n = dask.delayed(segment_chunk, nout=2)(
            input_block,
            channels,
            z_axis,
            diameter_yx,
            anisotropy,
            fast_mode,
            index,
            dim,
            do_3D,
            stitch_threshold,
            resample,
            cellprob_threshold,
            flow_threshold,
            model,
        )

        shape = input_block.shape[:-1]
        labeled_block = da.from_delayed(labeled_block, shape=shape, dtype=np.int32)

        #n = dask.delayed(np.int32)(n)
        n = da.from_delayed(n, shape=(), dtype=np.int32)

        total = n if total is None else total + n

        block_label_offset = da.where(labeled_block > 0, total, np.int32(0))
        labeled_block += block_label_offset

        labeled_blocks[index[:-1]] = labeled_block
        total += n

    # Put all the blocks together
    block_labeled = da.block(labeled_blocks.tolist())
    
    ## put the results of the chunks in a temporaray file, 
    ## otherwise it could recompute the chunks when doing the following function (post-processing to link the labels)
    ## Thanks Marvin for the idea !

    block_labeled = block_labeled.to_zarr(
       tmp_file,
       return_stored=False,
       compute=False,
       )

    _, total = dask.compute([block_labeled, total])[0]
    block_labeled = da.from_zarr(tmp_file)
    
    ## Post processing steps to link the labels between consecutives chunks using the overlaps
    depth = da.overlap.coerce_depth(len(depth), depth)

    if np.prod(block_labeled.numblocks) > 1:
        iou_depth = da.overlap.coerce_depth(len(depth), iou_depth)

        if any(iou_depth[ax] > depth[ax] for ax in depth.keys()):
            raise DistSegError("iou_depth (%s) > depth (%s)" % (iou_depth, depth))

        trim_depth = {k: depth[k] - iou_depth[k] for k in depth.keys()}
        block_labeled = da.overlap.trim_internal(
            block_labeled, trim_depth, boundary=boundary
        )
        block_labeled = link_labels(
            block_labeled,
            total,
            iou_depth,
            iou_threshold=iou_threshold,
        )

        block_labeled = da.overlap.trim_internal(
            block_labeled, iou_depth, boundary=boundary
        )

    else:
        block_labeled = da.overlap.trim_internal(
            block_labeled, depth, boundary=boundary
        )

    return block_labeled


def segment_chunk(
    chunk,
    channels,
    z_axis,
    diameter_yx,
    anisotropy,
    fast_mode,
    index,
    dim,
    do_3D,
    stitch_threshold, 
    resample,
    cellprob_threshold,
    flow_threshold,
    model,
):
    """Perform segmentation on an individual chunk."""
    # Cellpose seems to have some randomness, which is made deterministic by using the block
    # details as a random seed.
    np.random.seed(index)

    segments, _, _ = model.eval(
        chunk,
        channels=channels,
        z_axis=z_axis,
        channel_axis=dim-1,
        diameter=diameter_yx,
        do_3D=do_3D,
        stitch_threshold = stitch_threshold,
        cellprob_threshold = cellprob_threshold,
        flow_threshold = flow_threshold,
        resample=resample,
        anisotropy=anisotropy,
        augment=not fast_mode,
    )

    return segments.astype(np.int32), segments.max()


def link_labels(block_labeled, total, depth, iou_threshold=1):
    """
    Build a label connectivity graph that groups labels across blocks,
    use this graph to find connected components, and then relabel each
    block according to those.
    """
    label_groups = label_adjacency_graph(block_labeled, total, depth, iou_threshold)
    new_labeling = _label.connected_components_delayed(label_groups)
    return _label.relabel_blocks(block_labeled, new_labeling)


def label_adjacency_graph(labels, nlabels, depth, iou_threshold):
    all_mappings = [da.empty((2, 0), dtype=np.int32, chunks=1)]

    slices_and_axes = get_slices_and_axes(labels.chunks, labels.shape, depth)
    for face_slice, axis in slices_and_axes:
        face = labels[face_slice]
        mapped = _across_block_iou_delayed(face, axis, iou_threshold)
        all_mappings.append(mapped)

    i, j = da.concatenate(all_mappings, axis=1)
    result = _label._to_csr_matrix(i, j, nlabels + 1)
    return result


def _across_block_iou_delayed(face, axis, iou_threshold):
    """Delayed version of :func:`_across_block_label_grouping`."""
    _across_block_label_grouping_ = dask.delayed(_across_block_label_iou)
    grouped = _across_block_label_grouping_(face, axis, iou_threshold)
    return da.from_delayed(grouped, shape=(2, np.nan), dtype=np.int32)


def _across_block_label_iou(face, axis, iou_threshold):
    unique = np.unique(face)
    face0, face1 = np.split(face, 2, axis)

    intersection = sk_metrics.confusion_matrix(face0.reshape(-1), face1.reshape(-1))
    sum0 = intersection.sum(axis=0, keepdims=True)
    sum1 = intersection.sum(axis=1, keepdims=True)

    # Note that sum0 and sum1 broadcast to square matrix size.
    union = sum0 + sum1 - intersection

    # Ignore errors with divide by zero, which the np.where sets to zero.
    with np.errstate(divide="ignore", invalid="ignore"):
        iou = np.where(intersection > 0, intersection / union, 0)

    labels0, labels1 = np.nonzero(iou >= iou_threshold)

    labels0_orig = unique[labels0]
    labels1_orig = unique[labels1]
    grouped = np.stack([labels0_orig, labels1_orig])

    valid = np.all(grouped != 0, axis=0)  # Discard any mappings with bg pixels
    return grouped[:, valid]


def get_slices_and_axes(chunks, shape, depth):
    ndim = len(shape)
    depth = da.overlap.coerce_depth(ndim, depth)
    slices = da.core.slices_from_chunks(chunks)
    slices_and_axes = []
    for ax in range(ndim):
        for sl in slices:
            if sl[ax].stop == shape[ax]:
                continue
            slice_to_append = list(sl)
            slice_to_append[ax] = slice(
                sl[ax].stop - 2 * depth[ax], sl[ax].stop + 2 * depth[ax]
            )
            slices_and_axes.append((tuple(slice_to_append), ax))
    return slices_and_axes


"""
   To handle post pipeline analysis
   - Hierarchical clustering: from csv results file and segmented cells, perform and display clustering.

"""

import numpy as np
import pathlib, os, csv

import matplotlib as mpl
from matplotlib.backends.backend_qt5agg import FigureCanvas
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
        
from scipy.cluster.hierarchy import dendrogram
from scipy.cluster.hierarchy import fcluster
from scipy.spatial.distance import pdist, squareform 
from scipy.cluster.hierarchy import ward
from skimage.io import imsave
from sklearn.preprocessing import scale

import napari
from magicgui import magicgui
from napari.utils.notifications import show_info

import fish_feats.Utils as ut
import fish_feats.MainImage as mi
import fish_feats.FishWidgets as fwid
from qtpy.QtWidgets import QWidget, QVBoxLayout

## disable scipy cluster warning
from scipy.cluster.hierarchy import ClusterWarning
from warnings import simplefilter
simplefilter("ignore", ClusterWarning)

def do_hierarchy():
    mig = mi.MainImage( talkative=True )
    viewer = napari.current_viewer()
    viewer.title = "ZENnapari"
    
    filename = ut.dialog_filename()
    if filename is None:
        print("No file selected")
        return
    
    mig.open_image( filename=filename )
    ut.update_history(mig.imagedir)
    
    for chanel in range(mig.nbchannels):
        cmap = ut.colormapname(chanel)
        img = mig.get_channel(chanel)
        cview = viewer.add_image( img, name="originalChannel"+str(chanel), blending="additive", scale=(mig.scaleZ, mig.scaleXY, mig.scaleXY), colormap=cmap )
        dint = np.max(img)-np.min(img)
        cview.contrast_limits=(np.min(img), np.max(img)-0.75*dint)
    viewer.axes.visible = True
    get_scale = GetScales( viewer, mig )
    wid = viewer.window.add_dock_widget(get_scale, name="Scale")
    return wid

############ hierarchical analysis
class HierAnalysis:
    """ Perform and display hierarchical analysis based on selected features """

    def __init__(self):
        self.nclusters = 4
        self.wid = None
        self.clustward = None

    def set(self, mig, viewer):
        self.mig = mig
        self.viewer = viewer
        self.cluster_img = np.zeros(mig.get_image_shape(in2d=True), np.uint8)
        self.featlayer = self.viewer.add_labels(self.cluster_img, name="ClusteredCells", scale=(mig.scaleXY, mig.scaleXY), opacity=1)

    def get_data(self):
        """ Interface to select the file and the parameters """
        def load_file():
            """ Load the excel/csv file """
            with open(get_columns.load_file.value, 'r') as infile:
                csvreader = csv.DictReader(infile)
                print(csvreader.fieldnames)
                get_columns.use_column.choices = csvreader.fieldnames
        
        def load_table():
            """ Load the specific columns """
            keep = get_columns.use_column.value
            self.prepare_data( get_columns.load_file.value, keep )
            self.show_clusters()
        
        def update_clusters():
            """ Update all with the new number of clusters chosen """
            self.nclusters = int(get_columns.nb_clusters.value)
            self.show_clusters()

        def save_cluscells():
            """ Save image of cells colored by cluster """
            ccells = self.featlayer.data
            outname = self.mig.build_filename( endname="_ClusteredCells_nclus_"+str(self.nclusters)+".png" )
            vis = []
            for lay in self.viewer.layers:
                vis.append(lay.visible)
                lay.visible = False
            self.featlayer.visible = True
            screenshot = self.viewer.screenshot()
            for visib, lay in zip(vis, self.viewer.layers):
                lay.visible = visib

            imsave(outname, screenshot)
            show_info("Saved in "+outname)

        def save_dendrogram_img():
            """ Save image of dendrogram to file """
            outname = self.mig.build_filename(endname="_ClusterDendrogram_nclus_"+str(self.nclusters)+".png")
            self.fig.savefig(outname)
            show_info("Saved in "+outname)


        @magicgui(call_button="Cluster from selected columns", 
            use_column = dict(widget_type="Select", choices=[]), 
            nb_clusters={"widget_type": "Slider", "min":1, "max": 50},
            save_clustered_cells={"widget_type":"PushButton", "value": False},
            save_dendrogram={"widget_type":"PushButton", "value": False},
            )
        def get_columns( 
            #load_file=pathlib.Path(self.mig.rnacount_filename(ifexist=True)),
            load_file=pathlib.Path(self.mig.resdir),
            use_column = [],
            nb_clusters = 4,
            save_clustered_cells=False, save_dendrogram=False,
            ):
            load_table()
        
        get_columns.load_file.changed.connect(load_file)
        get_columns.nb_clusters.changed.connect(update_clusters)
        get_columns.save_clustered_cells.clicked.connect(save_cluscells)
        get_columns.save_dendrogram.clicked.connect(save_dendrogram_img)
        self.viewer.window.add_dock_widget( get_columns, name="Load data" )
        

    def get_cluster_colors(self):
        """ To have same color between the label layer and the matplotlib plot """
        return [0] + [mpl.colors.rgb2hex(self.featlayer.get_color(i+1)) for i in range(self.nclusters)]

    def check_label(self, columns, lab):
        """ Check if a label is in the list that should not """
        if lab in columns:
            ut.show_warning("Warning, "+lab+" is in the selected features list, that's weird")

    def prepare_data(self, filename, columns):
        """ normalisation of the data """
        res = []
        self.labels = []
        self.check_label(columns, "CellLabel")
        self.check_label(columns, "CellID")
        self.check_label(columns, "NucleusID")
        self.check_label(columns, "NucleusLabel")
        with open(filename, 'r') as infile:
            csvreader = csv.DictReader(infile)
            for row in csvreader:
                cres = []
                clab = int(row["CellLabel"])
                for col in columns:
                    cres.append(float(row[col])+1)
                res.append(cres)
                self.labels.append(clab)
            tab = np.array(res)    
            nans = np.isnan( res ).any( axis=1 )
            kinds = [ind for ind in range(tab.shape[0]) if not nans[ind] ] 
            tab = tab[ kinds ]
            self.labels = [ self.labels[ind] for ind in kinds]
            negs = ( tab<0 ).any( axis=1 )
            kinds = [ind for ind in range(tab.shape[0]) if not negs[ind] ] 
            tab = tab[ kinds ]
            self.labels = [ self.labels[ind] for ind in kinds]
            tab = np.log(tab)
            tab = scale(tab)
            print(tab.shape)

        dist_tab = pdist(tab, metric='euclidean')
        dist_tab = squareform(dist_tab)
        self.clustward = ward(dist_tab)

       
    def show_clusters(self):
        """ Show dendogram and classified cells """
        if self.clustward is None:
            return
        clustered = fcluster(self.clustward, t=self.nclusters, criterion="maxclust")
        
        self.mig.set_cells(self.cluster_img, clustered, self.labels)
        self.featlayer.refresh()
        #self.cmap = self.featlayer.colormap.colors 
        if self.wid is None:
            self.wid = self.create_plotwidget()
        self.update_plotwidget(clustered)
        if not ut.has_widget( self.viewer, "Dendrogram" ):
            self.viewer.window.add_dock_widget( self.wid, name="Dendrogram" )

    def create_plotwidget(self):
        mpl_widget = FigureCanvas( Figure(figsize=(6,6) ) )
        self.fig = mpl_widget.figure
        self.ax = mpl_widget.figure.subplots()
        return mpl_widget

    def update_plotwidget(self, clustered):
        clus_str = [ f"cluster #{l}: n={c}\n" for (l,c) in zip(*np.unique(clustered, return_counts=True)) ]

        cluster_colors = self.get_cluster_colors()
        cluster_colors_array = [cluster_colors[cl] for cl in clustered]
        link_cols = {}
        for i, i12 in enumerate(self.clustward[:,:2].astype(int)):
            c1, c2 = (link_cols[x] if x > len(self.clustward) else cluster_colors_array[x] for x in i12)
            link_cols[i+1+len(self.clustward)] = c1 if c1 == c2 else 'k'
    
        self.ax.cla()
        dend = dendrogram(self.clustward, p=8, truncate_mode='level', no_labels=True, ax=self.ax, link_color_func=lambda x: link_cols[x] )
        self.fig.canvas.draw_idle()
        #plt.show()


class GetScales( QWidget ):
    """ Interface to get metadata """

    def __init__(self, viewer, mig ):
        """ GUI """
        super().__init__()
        self.viewer = viewer
        self.mig = mig

        layout = QVBoxLayout()
        ## get scales parameters
        ## scalexy parameter
        scalexy_line, self.scale_xy = fwid.value_line( "Scale XY:", self.mig.scaleXY, descr="Pixel size in the lateral dimensions (XY), in microns" )
        layout.addLayout(scalexy_line)
        ## scalez parameter
        scalez_line, self.scale_z = fwid.value_line( "Scale Z:", self.mig.scaleZ, descr="Pixel size in the axial dimension (Z), in microns" )
        layout.addLayout(scalez_line)
        ## load filename
        load_line, self.segmented_cells = fwid.file_line( "Segmented cells file:", self.mig.junction_filename(dim=2,ifexist=True), "Select the segmented cells file to load", descr="File containing the segmented cells, should be a tif file with integer values" )
        layout.addLayout( load_line )
        
        ## btn go
        btn_update = fwid.add_button( "Update", self.update_scales, descr="Update the scale of the image and load the segmented cells", color=ut.get_color("go") )
        layout.addWidget(btn_update)

        self.setLayout(layout)

    def update_scales( self ):
        """ Go update scale """
        self.mig.scaleXY = float( self.scale_xy.text() )
        self.mig.scaleZ = float( self.scale_z.text() )
        for chan in range(self.mig.nbchannels):
            self.viewer.layers['originalChannel'+str(chan)].scale = [self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY]
        ut.remove_all_widget(self.viewer)
        self.mig.load_segmentation( self.segmented_cells.text() )
        done = self.mig.popFromJunctions()
        if done < 0:
            return
        
        hiera = HierAnalysis()
        hiera.set(self.mig, self.viewer)
        hiera.get_data()
    